@@ -3,7 +3,9 @@ from __future__ import annotations
 import abc
 from enum import Enum, EnumMeta
 from inspect import isclass
+import io
 import itertools
+import json
 from pathlib import Path
 import re
 from typing import Dict, Iterable, List, Optional, Tuple, Type, Union
@@ -38,7 +40,7 @@ class HppfclEnumMeta(EnumMeta):
 
 
 class GeometryType(Enum, metaclass=HppfclEnumMeta):
-    """Wraps GeometryTapes in an Enum.
+    """Wraps GeometryTypes in an Enum.
 
     Also contains the custom feature that it can be indexed by hppfcl Geometries due to inheriting from HppfclEnumMeta
     """
@@ -103,10 +105,10 @@ class Geometry(abc.ABC):
         The geometry class is the default way to describe a volume in space.
 
         :param parameters: Class-specific parameters defining the volume of the geometry. Will be unpacked in child
-          classes
+            classes
         :param pose: Transformation in world coordinates of this geometry
         :param hppfcl_representation: A hppfcl collision geometry representing this geometry in collision checks. If
-          not provided, will be created. If provided, there will be no checks whether it fits the given parameters!
+            not provided, will be created. If provided, there will be no checks whether it fits the given parameters!
         """
         assert self.type is not None, "Class {} needs to define a GeometryType".format(self.__class__)
         self.parameters = parameters
@@ -116,6 +118,10 @@ class Geometry(abc.ABC):
         else:
             self.collision_geometry = hppfcl_representation
 
+    def to_json_string(self) -> str:
+        """Returns a json string representation of this geometry."""
+        return json.dumps(self.serialized, indent=2)
+
     @classmethod
     def from_json_data(cls, description: Union[List, Dict[str, any]],
                        package_dir: Optional[Path] = None) -> Geometry:
@@ -148,6 +154,17 @@ class Geometry(abc.ABC):
         return class_ref(**desc)
 
     @classmethod
+    def from_json_string(cls, description: str, package_dir: Optional[Path] = None) -> Geometry:
+        """
+        Takes a serialized geometry specification and returns the according Geometry instance.
+
+        :param description: A serialized geometry
+        :param package_dir: The top directory to which the mesh file paths are relative to. Necessary for meshes (only).
+        """
+        return cls.from_json_data(json.loads(description), package_dir)
+
+    @classmethod
+    @abc.abstractmethod
     def from_hppfcl(cls, fcl: hppfcl.CollisionObject) -> Geometry:
         """
         A helper function that transforms a generic hppfcl collision object to an according Geometry instance.
@@ -155,10 +172,31 @@ class Geometry(abc.ABC):
         :param fcl: The hppfcl collision object
         :return: An instance of the according Obstacle class, based on the kind of collision object handed over.
         """
-        if cls is not Geometry:  # Enforce inheritance by children
-            raise NotImplementedError("from_hppfcl not implemented for class {}".format(cls))
 
-        return _geometry_type2class(GeometryType[type(fcl)]).from_hppfcl(fcl)
+    @classmethod
+    def from_pin_geometry(cls, geo: pin.GeometryObject) -> Geometry:
+        """
+        A helper function that transforms a generic pinocchio GeometryObject to an according Geometry instance.
+
+        :param geo: The pinocchio GeometryObject
+        :return: An instance of the according Obstacle class, based on the kind of collision object handed over.
+        """
+        if cls is not Geometry:
+            raise ValueError("Call from_pin_geometry on the Geometry class, not on a subclass.")
+        fcl = geo.geometry
+        will_be_instance_of = _geometry_type2class(GeometryType[type(fcl)])
+        intermediate = will_be_instance_of.from_hppfcl(hppfcl.CollisionObject(fcl, geo.placement.rotation,
+                                                                              geo.placement.translation))
+        if isinstance(intermediate, Mesh):
+            logging.debug("Restoring mesh file information from pinocchio geometry object")
+            intermediate.parameters['file'] = geo.meshPath
+            intermediate.parameters['scale'] = geo.meshScale
+        return intermediate
+
+    @property
+    def as_hppfcl_collision_object(self) -> Tuple[hppfcl.CollisionObject, ...]:
+        """Creates hppfcl CollisionObject(s) at the current pose of self."""
+        return hppfcl.CollisionObject(self.collision_geometry, self.placement.as_transform3f()),
 
     @property
     def collision_data(self) -> Tuple[Tuple[np.ndarray, hppfcl.CollisionGeometry], ...]:
@@ -219,7 +257,7 @@ class Box(Geometry):
     def from_hppfcl(cls, fcl: hppfcl.CollisionObject) -> Geometry:
         """Wraps a hppfcl Box in a Box instance."""
         x, y, z = [2 * hs for hs in fcl.collisionGeometry().halfSide]
-        return cls({'x': x, 'y': y, 'z': z}, fcl, fcl.getTransform())
+        return cls({'x': x, 'y': y, 'z': z}, fcl.getTransform(), fcl.collisionGeometry())
 
     def _make_collision_geometry(self) -> hppfcl.CollisionGeometry:
         """A hppfcl box is centered in the symmetry axes of the box defined by width, height, depth"""
@@ -279,7 +317,7 @@ class Cylinder(Geometry):
         """Wraps a hppfcl Cylinder in a Cylinder instance."""
         r = fcl.collisionGeometry().radius
         z = fcl.collisionGeometry().halfLength * 2
-        return cls({'r': r, 'z': z}, fcl, fcl.getTransform())
+        return cls({'r': r, 'z': z}, fcl.getTransform(), fcl.collisionGeometry())
 
     def _make_collision_geometry(self) -> hppfcl.CollisionGeometry:
         """
@@ -334,7 +372,7 @@ class Sphere(Geometry):
     def from_hppfcl(cls, fcl: hppfcl.CollisionObject) -> Geometry:
         """Wraps a hppfcl Sphere in a Sphere instance."""
         r = fcl.collisionGeometry().radius
-        return cls({'r': r}, fcl, fcl.getTransform())
+        return cls({'r': r}, fcl.getTransform(), fcl.collisionGeometry())
 
     def _make_collision_geometry(self) -> hppfcl.CollisionGeometry:
         """A hppfcl sphere is centered in the symmetry axes of the sphere defined by radius"""
@@ -382,12 +420,55 @@ class Mesh(Geometry):
     @classmethod
     def from_hppfcl(cls, fcl: hppfcl.CollisionObject) -> Geometry:
         """Wraps a hppfcl Mesh in a Mesh instance."""
-        return cls({'file': 0}, fcl, fcl.getTransform())  # TODO: File from hppfcl!
+        logging.debug("Creating mesh from hppfcl, will lose possible file location information")
+        return cls({'file': '', 'package_dir': ''}, fcl.getTransform(), fcl.collisionGeometry())
+
+    @staticmethod
+    def read_wrl(file: Path) -> Tuple[Tuple[np.ndarray, ...], Tuple[np.ndarray, ...]]:
+        """
+        Reads each sub-mesh from a wrl file into separate vertex and face arrays.
+
+        The WRL format describes triangulated meshes and can contain multiple sub meshes that are parsed separately.
+        We use these to store simplified convex decompositions of collision meshes.
+        They are created with https://github.com/gaschler/bounding-mesh
+
+        Based on example here: https://en.wikipedia.org/wiki/VRML
+        """
+        with open(file, "r") as f:
+            buffer = f.read()
+        if not buffer.startswith("#VRML V2.0 utf8"):
+            raise ValueError("wrl format not as expected")
+
+        shapes = re.findall(r'(?<=Shape {)[\n\t\w {.}\[\],-]*?(?=}\s*Shape|}\s*$)', buffer)
+        faces = tuple(re.findall(r'(?<=coordIndex \[)[0-9\s,-]*?(?=,?\s*])', s) for s in shapes)
+        vertices = tuple(re.findall(r'(?<=point \[)[e0-9.\s,-]*?(?=,?\s*])', s) for s in shapes)
+        if not all(len(f) == len(v) == 1 for f, v in zip(faces, vertices)):
+            raise ValueError("Every shape should have one face and vertex array.")
+
+        faces = tuple(np.genfromtxt(io.BytesIO(f.strip().replace("\t", "").replace(",\n", "\n").encode()),
+                                    dtype=int, delimiter=",")[:, :3] for f in itertools.chain(*faces))
+        vertices = tuple(np.genfromtxt(io.BytesIO(v.strip().replace("\t", "").replace(", \n", "\n").encode()),
+                                       dtype=float, delimiter=" ") for v in itertools.chain(*vertices))
+
+        if not all(np.all(f <= v.shape[0]) for f, v in zip(faces, vertices)):
+            raise ValueError("Faces should be valid indices in vertex")
+        return vertices, faces
 
     def _make_collision_geometry(self) -> hppfcl.CollisionGeometry:
-        """A hppfcl box is centered in the symmetry axes of the box defined by width, height, depth"""
-        mesh_loader = hppfcl.MeshLoader()
-        return mesh_loader.load(str(self.abs_filepath), self.scale)
+        """This method generates the hppfcl-based collision geometry for this mesh used for collision detection."""
+        if self.abs_filepath.suffix == ".wrl":
+            vertices, faces = self.read_wrl(self.abs_filepath)
+            geom = hppfcl.BVHModelOBBRSS()
+            num_tris = sum(f.shape[0] for f in faces)
+            geom.beginModel(num_tris, 3 * num_tris)
+            for vs, fs in zip(vertices, faces):
+                for f in fs:
+                    geom.addTriangle(vs[f[0], :], vs[f[1], :], vs[f[2], :])
+            geom.endModel()
+            return geom
+        else:
+            mesh_loader = hppfcl.MeshLoader()
+            return mesh_loader.load(str(self.abs_filepath), self.scale)
 
     @property
     def abs_filepath(self) -> Path:
@@ -473,6 +554,11 @@ class ComposedGeometry(Geometry):
         """This method is usually only called in the init, which is overwritten for this class."""
         raise NotImplementedError("ComposedGeometries collision objects are the sum of the internal geometries ones.")
 
+    @property
+    def as_hppfcl_collision_object(self) -> Tuple[hppfcl.CollisionObject, ...]:
+        """Creates hppfcl CollisionObjects at the current placement of self."""
+        return tuple(h for g in self.composing_geometries for h in g.as_hppfcl_collision_object)
+
     @property
     def collision_data(self) -> Tuple[Tuple[np.ndarray, hppfcl.CollisionGeometry], ...]:
         """Composed Geometries are the only ones with more than one element in the collision data"""
@@ -527,6 +613,11 @@ class EmptyGeometry(Geometry):
         parameters = dict()
         super().__init__(parameters, pose, hppfcl_representation)
 
+    @classmethod
+    def from_hppfcl(cls, fcl: hppfcl.CollisionObject) -> Geometry:
+        """This method is not implemented as it is not necessary."""
+        raise NotImplementedError("Empty geometries are not loaded from hppfcl.")
+
     @property
     def parameters(self) -> Dict[str, Union[float, str]]:
         """No Geometry, no parameters."""
@@ -537,6 +628,11 @@ class EmptyGeometry(Geometry):
         """No Geometry, no parameters."""
         pass
 
+    @property
+    def as_hppfcl_collision_object(self) -> Tuple[hppfcl.CollisionObject, ...]:
+        """An empty geometry does not have a collision object."""
+        return ()
+
     @property
     def collision_data(self) -> Tuple[Tuple[np.ndarray, hppfcl.CollisionGeometry], ...]:
         """No Geometry, no collision object."""
