@@ -1,6 +1,8 @@
 #!/usr/bin/env python3
 # Author: Jonathan Külz
 # Date: 12.01.22
+from __future__ import annotations
+
 import abc
 import itertools
 from pathlib import Path
@@ -17,7 +19,7 @@ import scipy.optimize as optimize
 
 from timor.task.Obstacle import Obstacle
 from timor.utilities import logging, spatial
-from timor.utilities.dtypes import float2array
+from timor.utilities.dtypes import IntermediateIkResult, float2array
 from timor.utilities.tolerated_pose import ToleratedPose
 from timor.utilities.transformation import Transformation, TransformationLike
 
@@ -88,7 +90,7 @@ class RobotBase(abc.ABC):
     # ---------- Static- and Classmethods ----------
     @classmethod
     @abc.abstractmethod
-    def from_urdf(cls, urdf_file: Path, package_dir: Path, **kwargs) -> 'RobotBase':
+    def from_urdf(cls, urdf_file: Path, package_dir: Path, **kwargs) -> RobotBase:
         """Class constructor from urdf file. kwargs should at least handle RobotBase keyword arguments."""
 
     # ---------- Abstract Properties ----------
@@ -164,7 +166,7 @@ class RobotBase(abc.ABC):
         """
 
     @abc.abstractmethod
-    def collisions(self, task: 'Task.Task') -> List[Tuple['RobotBase', Obstacle]]:  # noqa: F821
+    def collisions(self, task: 'Task.Task') -> List[Tuple[RobotBase, Obstacle]]:  # noqa: F821
         """
         Returns all collisions that appear between the robot and itself or obstacles in the task.
 
@@ -304,7 +306,7 @@ class RobotBase(abc.ABC):
 
         :param eef_pose: The desired 4x4 placement of the end effector
         :param q_init: The joint configuration to start with. If not given, will start the iterative optimization at the
-          current configuration
+            current configuration
         :return: A tuple of (q_solution, success [boolean])
         """
         return self.ik_scipy(eef_pose, q_init, **kwargs)
@@ -325,9 +327,9 @@ class RobotBase(abc.ABC):
         :param eef_pose: Desired end effector pose with tolerances.
         :param q_init: Initial joint angles to start searching from. If not given, uses current configuration
         :param max_iter: Maximum number of iterations before the algorithm gives up. If max_n_tries is larger than 1,
-          max_iter is counted over all n tries.
+            max_iter is counted over all n tries.
         :param restore_config: If True, the current robot configuration will be kept. If false, the found solution for
-          the inverse kinematics will be kept
+            the inverse kinematics will be kept
         :param max_n_tries: Maximum number of runs with a new random init configuration before the algorithm gives up
         :return: A tuple of q_solution [np.ndarray], success [boolean]
         """
@@ -447,7 +449,7 @@ class PinRobot(RobotBase):
 
     # ---------- Static- and Classmethods ----------
     @classmethod
-    def from_urdf(cls, urdf_file: Path, package_dir: Path, **kwargs) -> 'PinRobot':
+    def from_urdf(cls, urdf_file: Path, package_dir: Path, **kwargs) -> PinRobot:
         """Utility wrapper to load information about a robot from URDF and build the according PinRobot."""
         wrapper = pin.RobotWrapper.BuildFromURDF(str(urdf_file), str(package_dir))
         kwargs.setdefault('home_configuration', pin.neutral(wrapper.model))
@@ -488,8 +490,14 @@ class PinRobot(RobotBase):
 
     @property
     def mass(self) -> float:
-        """Robot mass in kg."""
-        return pin.computeTotalMass(self.model)
+        """
+        Robot mass in kg, included the unactuated mass belonging to the base.
+
+        Pinocchio calculates the robot mass as the actuated mass only, omitting the mass attached to the "universe".
+        Opposed to this approach, we include the mass of the (static) base in the robot mass.
+        :source: https://github.com/stack-of-tasks/pinocchio/issues/1286
+        """
+        return pin.computeTotalMass(self.model) + self.model.inertias[0].mass
 
     @property
     def njoints(self) -> int:
@@ -604,7 +612,7 @@ class PinRobot(RobotBase):
     def friction(self, dq: np.ndarray) -> np.ndarray:
         """Computes the forces due to friction"""
         # sin(atan( )) is smooth signum function
-        return self.model.damping*dq + self.model.friction*np.sin(np.arctan(100*dq))
+        return self.model.damping * dq + self.model.friction * np.sin(np.arctan(100 * dq))
 
     def motor_inertias(self, ddq: np.ndarray) -> np.ndarray:
         """Computes the motor inertias in the joins at a given acceleration"""
@@ -684,7 +692,7 @@ class PinRobot(RobotBase):
         :param friction: Whether to consider friction compensating torques
         :param eef_wrench: [f_x, f_y, f_z, tau_x, tau_y, tau_z] acting on the end-effector
         :param eef_wrench_frame: Frame the eef_wrench is described in (still acting on eef position!);
-          for now only "world" accepted. TODO : Extend with future frame definitions
+            for now only "world" accepted. TODO : Extend with future frame definitions
         """
         q = self.configuration if q is None else q
         dq = self.velocities if dq is None else dq
@@ -707,8 +715,8 @@ class PinRobot(RobotBase):
                 # force just needs rotation
                 last_joint_to_world[:3, :3] @ eef_wrench[:3],
                 # torque needs rotation + torques from offsetting force
-                last_joint_to_world[:3, :3] @ eef_wrench[3:] +
-                last_joint_to_world[:3, :3] @ (np.cross(last_joint_to_eef[:3, 3], eef_wrench[:3]))
+                last_joint_to_world[:3, :3] @ eef_wrench[3:]
+                + last_joint_to_world[:3, :3] @ (np.cross(last_joint_to_eef[:3, 3], eef_wrench[:3]))
             ))
             f_ext_vec.append(pin.Force(wrench_in_last_joint_frame))
         else:
@@ -728,7 +736,7 @@ class PinRobot(RobotBase):
 
         :param eef_pose: The desired 4x4 placement of the end effector
         :param q_init: The joint configuration to start with. If not given, will start the iterative optimization at the
-          current configuration
+            current configuration.
         :return: A tuple of (q_solution, success [boolean])
         """
         if self.njoints < 6:
@@ -748,21 +756,24 @@ class PinRobot(RobotBase):
         """
         Implements numerics inverse kinematics solvers based on the jacobian.
 
+        If no solution is found, returns success=False and the configuration q with the lowest euclidean distance to the
+        desired pose (ignoring orientation offset) that was found during max_iter iterations.
+
         Reference: Robotics, Siciliano et al. [2009], Chapter 3.7
 
         :param eef_pose: The desired 4x4 placement of the end effector
         :param q_init: The joint configuration to start with. If not given, will start the iterative optimization at the
-          current configuration
+            current configuration.
         :param gain: Gain Matrix K for closed "control" of q. Higher K leads to faster, but instable solutions.
         :param damp: Damping for the damped least squares pseudoinverse method
         :param max_iter: The maximum number of iterations before the algorithm fails.
-          (The default value 1000 is very conservative and can often be reduced by orders of magnitude;
-          run with logging set to debug to see how often ik fails due to hitting max iter for your problem)
+            (The default value 1000 is very conservative and can often be reduced by orders of magnitude;
+            run with logging set to debug to see how often ik fails due to hitting max iter for your problem)
         :param kind: One of "transpose", "pseudo_inverse", "damped_ls_inverse". The according variant of the Jacobian
-          will be used to solve the ik problem. While the transpose is the fastest one to calculate, for complicated
-          problems it is too inaccurate to converge (quickly) to a solution. Pseudo inverse and damped least squares
-          pseudo-inverse are pretty similar, the latter one introduces a "punishment" for the norm of the joint
-          velocities and is therefore more resistant against singularities.
+            will be used to solve the ik problem. While the transpose is the fastest one to calculate, for complicated
+            problems it is too inaccurate to converge (quickly) to a solution. Pseudo inverse and damped least squares
+            pseudo-inverse are pretty similar, the latter one introduces a "punishment" for the norm of the joint
+            velocities and is therefore more resistant against singularities.
         :return: A tuple of q_solution [np.ndarray], success [boolean]
         """
         # Custom errors to make sure no more deprecated arguments are used for the IK solver
@@ -774,7 +785,10 @@ class PinRobot(RobotBase):
 
         rot_join_mask = np.array([jnt.shortname() == 'JointModelRZ' for jnt in self.model.joints[1:]])
         if q_init is None:
-            q = self.configuration
+            # Find an initial guess for the solution where the robot is close to the desired point in space
+            candidates = [self.configuration] + [self.random_configuration() for _ in range(4)]
+            distances = [self.fk(cand).distance(eef_pose.nominal).translation_euclidean for cand in candidates]
+            q = candidates[np.argmin(distances)]
         else:
             q = q_init
         joint_idx_pin = self.model.frames[self.tcp].parent
@@ -793,16 +807,22 @@ class PinRobot(RobotBase):
 
         i = 0
         success = True
+        closest_translation = kwargs.get('closest_translation_q', IntermediateIkResult(q, -np.inf))
         while not eef_pose.valid(self.fk(q)):
             joint_current = self.data.oMi[joint_idx_pin]
             diff = joint_current.actInv(joint_desired)
             error_twist = pin.log(diff).vector
+            abs_translational_distance = np.linalg.norm(diff.translation)
+            if (abs_translational_distance > closest_translation.distance) and self.q_in_joint_limits(q) and \
+                    not self.has_self_collision(q):
+                closest_translation = IntermediateIkResult(q, abs_translational_distance)
             try:
                 J = pin.computeJointJacobian(self.model, self.data, q, joint_idx_pin)
                 J_inv = inv(J)  # Analytical Jacobian "pseudo inverse" (or transpose)
             except np.linalg.LinAlgError:
-                logging.debug(f"Jacobian ik break due to singularity after {i} iter")
-                return q, False
+                logging.debug(f"Jacobian ik break due to singularity after {i} iter for q={q}. Trying again.")
+                kwargs['closest_translation_q'] = closest_translation
+                return self.ik_jacobian(eef_pose, self.random_configuration(), gain, damp, max_iter - i, kind, **kwargs)
             q_dot = J_inv.dot(gain).dot(error_twist)
             q = pin.integrate(self.model, q, q_dot)
             # Python modolo defaults to positive values, but we want to preserve q sign
@@ -812,23 +832,28 @@ class PinRobot(RobotBase):
             q[rot_join_mask] = q[rot_join_mask] % (2 * np.pi) - sign_preserve[rot_join_mask]
 
             if np.inf in q:
-                logging.debug(f"Jacobian ik break due to q approaching inf after {i} iter")
-                return pin.neutral(self.model), False
+                logging.debug(f"Jacobian ik break due to q approaching inf after {i} iter. Trying again.")
+                kwargs['closest_translation_q'] = closest_translation
+                return self.ik_jacobian(eef_pose, self.random_configuration(), gain, damp, max_iter - i, kind, **kwargs)
             i += 1
             if i >= max_iter:
-                success = False
-                break
+                logging.debug(f"Jacobian ik not successful due to reaching the maximum number of iterations {i}")
+                logging.debug("Returning partial/closest solution.")
+                return closest_translation.q, False
 
         if not self.q_in_joint_limits(q):
             q[rot_join_mask] = spatial.rotation_in_bounds(q[rot_join_mask], self.joint_limits[:, rot_join_mask])
             if self.q_in_joint_limits(q):
                 # We're good to go
-                logging.debug("Successfully mapped ik result to (-pi, pi]")
+                logging.debug("Successfully mapped ik result to the robot joint limits")
             elif i < max_iter:
                 # Give it another try, starting from a different configuration
                 i += 1
                 logging.info("IK was out of joint limits, re-try")
+                kwargs['closest_translation_q'] = closest_translation
                 return self.ik_jacobian(eef_pose, self.random_configuration(), gain, damp, max_iter - i, kind, **kwargs)
+            else:
+                logging.debug("IK was out of joint limits, but we're out of iterations")
 
         if self.has_self_collision(q):
             logging.info("IK solution had self-collisions, re-try")
@@ -837,22 +862,22 @@ class PinRobot(RobotBase):
             if i < max_iter:
                 # Give it another try, starting from a different configuration
                 i += 1
+                kwargs['closest_translation_q'] = closest_translation
                 return self.ik_jacobian(eef_pose, self.random_configuration(), gain, damp, max_iter - i, kind, **kwargs)
+            logging.debug("Jacobian ik not successful after maximum number of iterations, fails with self collision.")
             success = False
-
-        if 'task' in kwargs and self.has_collisions(kwargs['task']):
+        elif 'task' in kwargs and self.has_collisions(kwargs['task']):
             logging.info("IK solution collides with environment, re-try")
             if i < max_iter:
                 # Give it another try, starting from a different configuration
                 i += 1
+                logging.debug("Keep the previously best q, but without checking for environment collisions.")
+                kwargs['closest_translation_q'] = closest_translation
                 return self.ik_jacobian(eef_pose, self.random_configuration(), gain, damp, max_iter - i, kind, **kwargs)
+            logging.debug("Jacobian ik not successful after maximum number of iterations, fails due to collision.")
             success = False
 
-        if success:
-            logging.debug(f"Jacobian ik successful after {i} iterations")
-        else:
-            logging.debug(f"Jacobian ik not successful due to reaching the maximum number of iterations {i}")
-
+        logging.debug(f"Jacobian ik successful after {i} iterations")
         return q, success
 
     def move(self, displacement: TransformationLike):
@@ -888,14 +913,16 @@ class PinRobot(RobotBase):
         try:
             return pin.randomConfiguration(self.model)
         except RuntimeError:
-            logging.info('Random configuration failed due to infinite joint limits - setting to (-2pi, 2pi)')
+            logging.debug('Random configuration failed due to infinite joint limits - setting to (-2pi, 2pi)')
             lower = np.maximum(self.joint_limits[0, :], -2 * np.pi)
             upper = np.minimum(self.joint_limits[1, :], 2 * np.pi)
             return (upper - lower) * np.random.rand(self.njoints) + lower
 
     def set_base_placement(self, placement: TransformationLike) -> None:
         """Moves the base to desired position"""
-        self.move(self.placement.inv.multiply_from_left(placement))
+        placement = Transformation(placement)
+        if placement != self._base_placement:
+            self.move(placement @ self.placement.inv)
 
     def static_torque(self, q: np.ndarray, f_ext: pin.StdVec_Force = None) -> np.ndarray:
         """
@@ -959,7 +986,7 @@ class PinRobot(RobotBase):
 
         :param task: The task to check for collisions
         :param return_at_first: If true, this function behaves like a "has_collision". If false, it behaves like a
-          "get_all_collisions" function,
+            "get_all_collisions" function.
         :return: Bool if return_at_first is true, otherwise a list of all colliding pairs.
         """
         collisions: List[Tuple[RobotBase, Union[RobotBase, Obstacle]]] = list()
@@ -975,8 +1002,9 @@ class PinRobot(RobotBase):
         for obstacle in task.obstacles:
             for (robot_geometry, omg), (t, geometry) in \
                     itertools.product(zip(geometries, self.collision_data.oMg), obstacle.collision.collision_data):
+                omg3f = Transform3f(omg.homogeneous[:3, :3], omg.homogeneous[:3, 3])
                 t3f = Transform3f(t[:3, :3], t[:3, 3])
-                if collide(robot_geometry, omg, geometry, t3f, request, result):
+                if collide(robot_geometry, omg3f, geometry, t3f, request, result):
                     if return_at_first:
                         return True
                     collisions.append((self, obstacle))
@@ -1066,7 +1094,7 @@ class PinRobot(RobotBase):
 
         :param visualizer: A meshcat visualizer instance. If none, a new will be created
         :param coordinate_systems: Can be None, 'tcp', 'joints' or 'full' - the specified coordinates will be drawn
-          (full means all frames, as in fk)
+            (full means all frames, as in fk).
         :return: The meshcat visualizer instance used for plotting the robot
         """
         if coordinate_systems not in (None, 'joints', 'full', 'tcp'):
@@ -1217,23 +1245,24 @@ class PinRobot(RobotBase):
         :param parent: Either the index or the name of the parent joint to which the new one shall be appended
         :param bodies: All bodies attached to the joint. To add N bodies, there need to be N-1 fixed joints.
         :param fixed_joints: Fixed joints attached to the joint. Assumes bodies and fixed joints are added in the order
-          b1-fj1-b2-fx2-...-bn
+            b1-fj1-b2-fx2-...-bn
         :param inertia: The inertia attached to the joint. If not provided, will be auto-calculated by the bodies added.
         :param update_kinematics: If False, the datas will not be re-calculated, meaning the changes to not have
-          an immediate effect on the results of kinematic/dynamic methods. Is useful when adding multiple joints,
-          so just after the last change the data needs to be updated.
+            an immediate effect on the results of kinematic/dynamic methods. Is useful when adding multiple joints,
+            so just after the last change the data needs to be updated.
         :param update_home_collisions: If False, the visual and collision geometries will not be updated,
-          meaning this joint will neither be plotted nor considered for collision checking until done so.
-          Useful for either building models that don't need either of these features or when adding multiple joints,
-          so this can be done after the last only.
+            meaning this joint will neither be plotted nor considered for collision checking until done so.
+            Useful for either building models that don't need either of these features or when adding multiple joints,
+            so this can be done after the last only.
         :param kwargs: All key-word-arguments supported by pin.Model.addJoint:
-            - joint_placement: [SE3] -> Placement of the joint inside its parent joint.
-            - joint_name: [str] -> If empty, name will be random
-            - max_effort: [float] -> Maximal Joint Torque
-            - max_velocity: [float] -> Maximal Joint Velocity
-            - min_config and max_config: [float] -> Joint configuration limits
-            - friction: [float] -> Joint friction parameters
-            - damping: [float] -> Joint damping parameters
+
+            * joint_placement: [SE3] -> Placement of the joint inside its parent joint.
+            * joint_name: [str] -> If empty, name will be random
+            * max_effort: [float] -> Maximal Joint Torque
+            * max_velocity: [float] -> Maximal Joint Velocity
+            * min_config and max_config: [float] -> Joint configuration limits
+            * friction: [float] -> Joint friction parameters
+            * damping: [float] -> Joint damping parameters
         :return: Index of the new joint
         """
         if len(fixed_joints) > 0:
