@@ -1,7 +1,14 @@
 from __future__ import annotations
-from dataclasses import dataclass, field
+
+from collections import namedtuple
+from dataclasses import asdict, dataclass, field, fields
+import datetime
+import logging
+from pathlib import Path
 import random
-from typing import Callable, Collection, Dict, Generator, Iterable, List, Tuple, Union
+import re
+from typing import Any, Callable, Collection, Dict, Generator, Generic, Iterable, List, Tuple, Type, TypeVar, Union, \
+    get_type_hints
 
 from hppfcl import hppfcl
 from matplotlib import pyplot as plt
@@ -10,7 +17,11 @@ import pinocchio as pin
 from roboticstoolbox.tools.trajectory import mstraj
 
 import timor.utilities.errors as err
-from timor.utilities.transformation import TransformationLike, Transformation
+from timor.utilities.schema import DEFAULT_DATE_FORMAT
+from timor.utilities.transformation import Transformation, TransformationLike
+
+T = TypeVar('T')
+IntermediateIkResult = namedtuple("IntermediateIkResult", ["q", "distance"])
 
 
 class EternalDict(dict):
@@ -36,15 +47,15 @@ class EternalDict(dict):
     popitem = _immutable
 
 
-class Lazy:
+class Lazy(Generic[T]):
     """An implementation of lazy variables in python"""
 
-    def __init__(self, func):
+    def __init__(self, func: Callable[[], T]):
         """Initialize an "empty" lazy variable with the function to be evaluated lazily"""
-        self.func = func
+        self.func: Callable[[], T] = func
         self.value = None
 
-    def __call__(self, *args, **kwargs):
+    def __call__(self, *args, **kwargs) -> T:
         """Useful s.t. the variable can be evaluated like lazy()"""
         return self.value_or_func()
 
@@ -68,7 +79,7 @@ class Lazy:
         else:
             return other * self()
 
-    def __getstate__(self):
+    def __getstate__(self) -> T:
         """Lazy variables will always be evaluated when serialized"""
         return self.value_or_func()
 
@@ -76,14 +87,14 @@ class Lazy:
         """Reset the value of the lazy variable s.t. it must be re-evaluated next time the value is requested"""
         self.value = None
 
-    def value_or_func(self):
+    def value_or_func(self) -> T:
         """Evaluates the lazy variable. Is triggered by self()"""
         if self.value is None:
             self.value = self.func()
         return self.value
 
     @classmethod
-    def from_variable(cls, variable) -> 'Lazy':
+    def from_variable(cls, variable: any) -> Lazy:
         """Useful wrapper to prevent evaluation of Lazy variables when multiplied with another variable"""
 
         def ret():
@@ -124,21 +135,21 @@ class SingleSet(set):
             raise err.UniqueValueError(self._err.format(element))
         super().add(element)
 
-    def copy(self) -> 'SingleSet':
+    def copy(self) -> SingleSet:
         """Custom copy should also return a SingleSet"""
         return self.__class__(super().copy())
 
-    def filter(self, func: Callable[[any], bool]):
+    def filter(self, func: Callable[[any], bool]) -> SingleSet:
         """
         Convenience function to avoid set comparison over and over again
 
         :param func: A filter function that takes an element of this set and evaluates to true if
-          the element should be kept
+            the element should be kept
         :return: A new set of elements fulfilling the filter criterion
         """
         return self.__class__(element for element in self if func(element))
 
-    def intersection(self, *s: Iterable) -> 'SingleSet':
+    def intersection(self, *s: Iterable) -> SingleSet:
         """Intersection should also retutrn a SingleSet"""
         return self.__class__(element for arg in s for element in arg if element in self)
 
@@ -155,7 +166,7 @@ class SingleSet(set):
                 self.add(element)
                 added.add(element)
 
-    def union(self, *s: Iterable) -> 'SingleSet':
+    def union(self, *s: Iterable) -> SingleSet:
         """Returns a SingleSet of the union of this and the input"""
         elements = self.__class__()  # This is needed so this classes __contains__ method is used on the input as well
         for arg in s:
@@ -168,6 +179,78 @@ class SingleSet(set):
         return self.__class__(super().union(elements))
 
 
+@dataclass
+class TypedHeader:
+    """A dataclass that ensures and transforms {Module, Task, Solution}-header data to the desired datatypes."""
+
+    def __post_init__(self):
+        """Post init function that ensures the correct datatypes"""
+        self.__setattr__('__immutable', False)
+        for name, type_annotation in get_type_hints(self, globalns=globals(), localns=locals()).items():
+            value = self.__getattribute__(name)
+            self.__setattr__(name, self.cast(value, type_annotation))
+        self.__setattr__('__immutable', True)
+
+    @staticmethod
+    def cast(value: Any, cast_to_type: Type) -> Any:
+        """Casts the value to the dtype"""
+        if cast_to_type is datetime.datetime:
+            return TypedHeader.cast_to_datetime(value)
+        try:
+            return cast_to_type(value)
+        except TypeError:
+            logging.debug("Could not cast {} to type {}. Returning original value.".format(value, cast_to_type))
+            return value
+
+    @staticmethod
+    def cast_to_datetime(value: Union[str, datetime.datetime]) -> datetime.datetime:
+        """Casts the value to a datetime.datetime object"""
+        if isinstance(value, datetime.datetime):
+            return value
+        if isinstance(value, str):
+            value = re.sub('/', '-', value)  # Allows parsing yyyy/mm/dd formatted dates
+            return datetime.datetime.strptime(value, DEFAULT_DATE_FORMAT)
+        raise TypeError(f"Cannot cast {value} to datetime.datetime")
+
+    @staticmethod
+    def string_list_factory() -> field:
+        """Returns a field with a default factory that returns a list with one empty string"""
+
+        def list_factory() -> List[str]: return ['']
+
+        return field(default_factory=list_factory)
+
+    @classmethod
+    def fields(cls) -> Tuple[str, ...]:
+        """Returns the fields of this class"""
+        return tuple(f.name for f in fields(cls))
+
+    def asdict(self) -> Dict[str, Any]:
+        """Returns a dictionary representation of this object"""
+        return asdict(self)
+
+    def _raise_immutable(self, f: Callable) -> Callable:
+        """Raises error if objects in this dataclass are to be resetted but the instance is set to immutable"""
+
+        def wrapper(*args, **kwargs):
+            try:
+                if self.__getattribute__('__immutable'):
+                    raise TypeError("Cannot reset attributes of an immutable object")
+            except AttributeError:  # __init__, __immutable not yet set
+                pass
+            return f(*args, **kwargs)
+
+        return wrapper
+
+    def __delattr__(self, item):
+        """Delete attribute"""
+        self._raise_immutable(super().__delattr__)(item)
+
+    def __setattr__(self, key, value):
+        """Sets an attribute of this class"""
+        return self._raise_immutable(super().__setattr__)(key, value)
+
+
 @dataclass
 class TorqueInput:
     """Holds data defining robot control parameters"""
@@ -238,34 +321,6 @@ class Trajectory:
             ddq = (self.dq[1:] - self.dq[:-1]) / time_delta[:, None]
             self.ddq = np.vstack([init, ddq])
 
-    def plot(self, show_goals_reached: bool = False) -> plt.Figure:
-        """
-        Plot this trajectory
-
-        :param show_goals_reached: If set to true, the time at which goals are reached is highlighted in the plot
-        :return: Figure with a plot of this trajectory.
-        """
-        f, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex='all')
-        ax1.plot(self.t, self.q)
-        ax2.plot(self.t, self.dq)
-        ax3.plot(self.t, self.ddq)
-
-        ax1.set(ylabel="Configuration q")
-        ax2.set(ylabel="Velocity dq")
-        ax3.set(xlabel="Time [s]", ylabel="Acceleration ddq")
-
-        if show_goals_reached:
-            t_offset = 0.01 * (max(self.t) - min(self.t))
-            for goal_id, goal_time in self.goals.items():
-                for ax in (ax1, ax2, ax3):
-                    ax.axvline(goal_time, c='blue')
-                    ax.text(goal_time + t_offset, .9, goal_id, c='blue', transform=ax.get_xaxis_transform())
-
-        plt.tight_layout()
-        plt.show()
-
-        return f
-
     @staticmethod
     def from_mstraj(via_points: np.array, dt: float, t_acc: float, qd_max: float, **kwargs) -> 'Trajectory':
         """
@@ -299,11 +354,51 @@ class Trajectory:
         ddq = np.zeros_like(q)
         return Trajectory(t=np.asarray([0, time]), q=q, dq=dq, ddq=ddq, goals={})
 
+    def to_json_data(self) -> Dict:
+        """Return trajectory as a dictionary that can be serialized as json."""
+        return {
+            "t": self.t.tolist(),
+            "q": self.q.tolist(),
+            "dq": self.dq.tolist(),
+            "ddq": self.ddq.tolist(),
+            "goal2time": self.goals
+        }
+
+    def plot(self, show_goals_reached: bool = False) -> plt.Figure:
+        """
+        Plot this trajectory
+
+        :param show_goals_reached: If set to true, the time at which goals are reached is highlighted in the plot
+        :return: Figure with a plot of this trajectory.
+        """
+        f, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex='all')
+        ax1.plot(self.t, self.q)
+        ax2.plot(self.t, self.dq)
+        ax3.plot(self.t, self.ddq)
+
+        ax1.set(ylabel="Configuration q")
+        ax2.set(ylabel="Velocity dq")
+        ax3.set(xlabel="Time [s]", ylabel="Acceleration ddq")
+
+        if show_goals_reached:
+            t_offset = 0.01 * (max(self.t) - min(self.t))
+            for goal_id, goal_time in self.goals.items():
+                for ax in (ax1, ax2, ax3):
+                    ax.axvline(goal_time, c='blue')
+                    ax.text(goal_time + t_offset, .9, goal_id, c='blue', transform=ax.get_xaxis_transform())
+
+        plt.tight_layout()
+        plt.show()
+
+        return f
+
     def __add__(self, other: 'Trajectory'):
         """Append another trajectory to self"""
         if not isinstance(other, Trajectory):
             return NotImplemented
-        if self.t.size > 1:
+        if (other.t.size == 0) or (self.t.size == 0):
+            return self
+        elif self.t.size > 1:
             # dt is the offset between end of self and start of other
             dt = self.t[-1] - self.t[-2]
         else:
@@ -316,7 +411,7 @@ class Trajectory:
         t_new = np.hstack((self.t, other.t + offset_other))
         q_new = np.vstack((self.q, other.q))
         # Offset the second trajectories goals
-        goals_new = self.goals | {k: v + offset_other for k, v in other.goals.items()}
+        goals_new = {**self.goals, **{k: v + offset_other for k, v in other.goals.items()}}
 
         dq_new = np.vstack((self.dq, other.dq))
         ddq_new = np.vstack((self.ddq, other.ddq))
@@ -333,6 +428,19 @@ class Trajectory:
             ddq=self.ddq[item]
         )
 
+    def __eq__(self, other) -> bool:
+        """Test if two trajectories are close together (max distance < 1e-5)"""
+        if type(self) is type(other):
+            return all((np.allclose(self.t, other.t),
+                        np.allclose(self.q, other.q),
+                        np.allclose(self.dq, other.dq),
+                        np.allclose(self.ddq, other.ddq),
+                        self.goals.keys() == other.goals.keys(),
+                        all(abs(self.goals[k] - other.goals[k]) < 1e-5 for k in self.goals.keys())))
+        elif isinstance(other, Trajectory):
+            raise NotImplementedError(f"Comparison of {type(self)} and {type(other)} not yet implemented.")
+        return NotImplemented
+
 
 def deep_iter_dict(d: Dict[any, any], max_depth=10, _prev=None) -> Generator[Tuple[List[any], any], None, None]:
     """
@@ -342,8 +450,9 @@ def deep_iter_dict(d: Dict[any, any], max_depth=10, _prev=None) -> Generator[Tup
     :param max_depth: The max recursion depth to dive into the dictionary
     :param _prev: Previously seen keys. Usually would not be set by the user.
     :return: Yields a generator that returns a tuple of:
-        - The keys leading to this values in a list like: [key_lvl_1, key_lvl_2, ..., key_lvl_n]
-        - The value if it is not yet another dictionary
+
+        * The keys leading to this values in a list like: [key_lvl_1, key_lvl_2, ..., key_lvl_n]
+        * The value if it is not yet another dictionary
     """
     if _prev is None:
         _prev = []
@@ -353,7 +462,7 @@ def deep_iter_dict(d: Dict[any, any], max_depth=10, _prev=None) -> Generator[Tup
         if not isinstance(val, dict):
             yield _prev + [key], val
         else:
-            yield from deep_iter_dict(val, max_depth-1, _prev + [key])
+            yield from deep_iter_dict(val, max_depth - 1, _prev + [key])
 
 
 def float2array(scalar: float) -> np.ndarray:
@@ -387,7 +496,7 @@ def fuzzy_dict_key_matching(dict_in: Dict[str, any],
     :param dict_in: An input dictionary that shall be "fuzzy-matched". The keys must be strings.
     :param aliases: A mapping from "valid alias key in dict_in" -> "valid real key"
     :param desired_only: If given, the returned dictionary will only contain this subset of keys.
-      Useful for discarding aliases after matching
+        Useful for discarding aliases after matching
     :param ignore_casing: If true, keys will be matched case-insensitive to aliases AND TO KEYS FROM DESIRED ONLY
     :return: The resulting dictionary containing all or only the desired mappings, taking aliases into account
     """
@@ -395,9 +504,11 @@ def fuzzy_dict_key_matching(dict_in: Dict[str, any],
         aliases = {}
     if ignore_casing:
         # Define the operation performed on
-        def strfnc(s: str): return s.lower()
+        def strfnc(s: str):
+            return s.lower()
     else:
-        def strfnc(s: str): pass
+        def strfnc(s: str):
+            pass
     if desired_only is not None:
         # Add them to aliases to possibly enable string-insensitive matching for desired keys
         aliases.update({desired: desired for desired in desired_only})
@@ -434,6 +545,13 @@ def hppfcl_collision2pin_geometry(geometry: hppfcl.CollisionGeometry,
         return pin.GeometryObject(name, parent_frame, parent_joint, geometry, placement)
 
 
+def map2path(p: Union[str, Path]) -> Path:
+    """Maps strings to paths, but does nothing else s.t. e.g. Django paths are not casted."""
+    if isinstance(p, str):
+        return Path(p)
+    return p
+
+
 def possibly_nest_as_list(value: any, tuple_ok: bool = True):
     """
     A utility to wrap a value within a list if it is not already.
